import sys

import numpy as np
import pytest

import rlic

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup

img = u = v = np.eye(64)
kernel = np.linspace(0, 1, 10, dtype="float64")


def assert_exceptions_match(
    to_eval: str,
    global_namespace,
    local_namespace,
    group_match: str,
    expected_sub: list[tuple[type[Exception], str]],
) -> None:
    with pytest.raises(ExceptionGroup, match=group_match) as excinfo:
        eval(to_eval, global_namespace, local_namespace)

    for exctype, match in expected_sub:
        assert excinfo.group_contains(exctype, match=match, depth=1)

    assert len(excinfo.value.exceptions) == len(expected_sub)


def test_invalid_iterations():
    with pytest.raises(
        ValueError,
        match=(
            r"^Invalid number of iterations: -1\n"
            r"Expected a strictly positive integer\.$"
        ),
    ):
        rlic.convolve(img, u, v, kernel=kernel, iterations=-1)


def test_invalid_uv_mode():
    with pytest.raises(
        ValueError,
        match=(
            r"^Invalid uv_mode 'astral'\. Expected one of \['velocity', 'polarization'\]$"
        ),
    ):
        rlic.convolve(img, u, v, kernel=kernel, uv_mode="astral")


def test_invalid_texture_ndim():
    img = np.ones((16, 16, 16))
    assert_exceptions_match(
        "rlic.convolve(img, u, v, kernel=kernel)",
        globals(),
        locals(),
        group_match=r"^Invalid inputs were received\.",
        expected_sub=[
            (
                ValueError,
                r"^Expected a texture with exactly two dimensions\. Got texture\.ndim=3$",
            ),
            (
                ValueError,
                r"^Shape mismatch: expected texture, u and v with identical shapes\.",
            ),
        ],
    )


def test_invalid_texture_shape_and_ndim():
    img = np.ones((16, 16, 16))

    assert_exceptions_match(
        "rlic.convolve(img, u, v, kernel=kernel)",
        globals(),
        locals(),
        group_match=r"^Invalid inputs were received\.",
        expected_sub=[
            (
                ValueError,
                r"^Expected a texture with exactly two dimensions\. Got texture\.ndim=3$",
            ),
            (
                ValueError,
                r"^Shape mismatch: expected texture, u and v with identical shapes\.",
            ),
        ],
    )


def test_invalid_texture_values():
    img = -np.ones((64, 64))
    with pytest.raises(
        ValueError,
        match=(
            r"^Found invalid texture element\(s\)\. Expected only positive values\.$"
        ),
    ):
        rlic.convolve(img, v, v, kernel=kernel)


@pytest.mark.parametrize(
    "texture_shape, u_shape, v_shape",
    [
        ((64, 64), (65, 64), (64, 64)),
        ((64, 64), (64, 64), (63, 64)),
        ((64, 66), (64, 64), (64, 64)),
    ],
)
def test_mismatched_shapes(texture_shape, u_shape, v_shape):
    prng = np.random.default_rng(0)
    texture = prng.random(texture_shape)
    u = prng.random(u_shape)
    v = prng.random(v_shape)
    with pytest.raises(
        ValueError,
        match=(
            r"^Shape mismatch: expected texture, u and v with identical shapes\. "
            rf"Got texture.shape=\({texture.shape[0]}, {texture.shape[1]}\), "
            rf"u.shape=\({u.shape[0]}, {u.shape[1]}\), "
            rf"v.shape=\({v.shape[0]}, {v.shape[1]}\)$"
        ),
    ):
        rlic.convolve(texture, u, v, kernel=kernel)


def test_invalid_kernel_ndim():
    with pytest.raises(
        ValueError,
        match=r"^Expected a kernel with exactly one dimension\. Got kernel\.ndim=2$",
    ):
        rlic.convolve(img, u, v, kernel=np.ones((5, 5)))


@pytest.mark.parametrize("polluting_value", [-np.inf, np.inf, np.nan])
def test_non_finite_kernel(polluting_value):
    kernel = np.ones(11)
    kernel[5] = polluting_value
    with pytest.raises(
        ValueError,
        match=r"^Found non-finite value\(s\) in kernel\.$",
    ):
        rlic.convolve(img, u, v, kernel=kernel)


def test_invalid_texture_dtype():
    img = np.ones((64, 64), dtype="complex128")
    assert_exceptions_match(
        "rlic.convolve(img, u, v, kernel=kernel)",
        globals(),
        locals(),
        group_match=r"^Invalid inputs were received\.",
        expected_sub=[
            (
                TypeError,
                r"^Found unsupported data type\(s\): \[dtype\('complex128'\)\]\. "
                r"Expected texture, u, v and kernel with identical dtype, from "
                r"\[dtype\('float32'\), dtype\('float64'\)\]\. "
                r"Got texture\.dtype=dtype\('complex128'\), u\.dtype=dtype\('float64'\), "
                r"v\.dtype=dtype\('float64'\), kernel\.dtype=dtype\('float64'\)$",
            ),
            (TypeError, r"^Data types mismatch"),
        ],
    )


def test_invalid_kernel_dtype():
    assert_exceptions_match(
        "rlic.convolve(img, u, v, kernel=-np.ones(5, dtype='complex128'))",
        globals(),
        locals(),
        group_match=r"^Invalid inputs were received\.",
        expected_sub=[
            (
                TypeError,
                r"^Found unsupported data type\(s\): \[dtype\('complex128'\)\]\. "
                r"Expected texture, u, v and kernel with identical dtype, from "
                r"\[dtype\('float32'\), dtype\('float64'\)\]\. "
                r"Got texture\.dtype=dtype\('float64'\), u\.dtype=dtype\('float64'\), "
                r"v\.dtype=dtype\('float64'\), kernel\.dtype=dtype\('complex128'\)$",
            ),
            (TypeError, r"^Data types mismatch"),
        ],
    )


def test_mismatched_dtypes():
    img = np.ones((64, 64), dtype="float32")
    with pytest.raises(
        TypeError,
        match=(
            r"^Data types mismatch. "
            r"Expected texture, u, v and kernel with identical dtype, from "
            r"\[dtype\('float32'\), dtype\('float64'\)\]\. "
            r"Got texture\.dtype=dtype\('float32'\), u\.dtype=dtype\('float64'\), "
            r"v\.dtype=dtype\('float64'\), kernel\.dtype=dtype\('float64'\)$"
        ),
    ):
        rlic.convolve(img, u, v, kernel=kernel)


def test_all_validators_before_returns():
    # until v0.3.2, iterations=0 implied an early return that skipped
    # most validators.
    kernel = np.full(11, np.nan)
    with pytest.raises(
        ValueError,
        match=r"^Found non-finite value\(s\) in kernel\.$",
    ):
        rlic.convolve(img, u, v, kernel=kernel, iterations=0)



import sys
from importlib.metadata import version
from importlib.util import find_spec


def pytest_report_header(config, start_path):
    is_gil_enabled = sys.version_info < (3, 13) or sys._is_gil_enabled()

    return [
        f"{is_gil_enabled = }",
        f"NumPy: {version('numpy')}",
        f"rlic._core loads from {find_spec('rlic._core').origin}",
    ]



from __future__ import annotations

__all__ = ["convolve"]

import sys
from typing import TYPE_CHECKING, Literal

import numpy as np

from rlic._core import convolve_f32, convolve_f64

if sys.version_info < (3, 11):
    from exceptiongroup import ExceptionGroup  # pyright: ignore[reportUnreachable]

if TYPE_CHECKING:
    from typing import TypeVar

    from numpy import dtype, ndarray
    from numpy import float32 as f32
    from numpy import float64 as f64

    F = TypeVar("F", f32, f64)

_KNOWN_UV_MODES = ["velocity", "polarization"]
_SUPPORTED_DTYPES: list[np.dtype[np.floating]] = [
    np.dtype("float32"),
    np.dtype("float64"),
]


def convolve(
    texture: ndarray[tuple[int, int], dtype[F]],
    /,
    u: ndarray[tuple[int, int], dtype[F]],
    v: ndarray[tuple[int, int], dtype[F]],
    *,
    kernel: ndarray[tuple[int], dtype[F]],
    uv_mode: Literal["velocity", "polarization"] = "velocity",
    iterations: int = 1,
) -> ndarray[tuple[int, int], dtype[F]]:
    """2-dimensional line integral convolution.

    Apply Line Integral Convolution to a texture array, against a 2D flow (u, v)
    and via a 1D kernel.

    Arguments
    ---------
    texture: 2D numpy array, positional-only
      Think of this as a tracer fluid. Random noise is a good input in the
      general case.

    u, v: 2D numpy arrays
      Represent the horizontal and vertical components of a vector field,
      respectively.

    kernel: 1D numpy array, keyword-only
      This is the convolution kernel. Think of it as relative weights along a
      portion of a field line. The first half of the array represent weights on
      the "past" part of a field line (with respect to a starting point), while
      the second line represents weights on the "future" part.

    uv_mode: 'velocity' (default), or 'polarization', keyword-only
      By default, the vector (u, v) field is assumed to be velocity-like, i.e.,
      its direction matters. With uv_mode='polarization', direction is
      effectively ignored.

    iterations: (positive) int (default: 1), keyword-only
      Perform multiple iterations in a loop where the output array texture is
      fed back as the input to the next iteration. Looping is done at the
      native-code level.

    Returns
    -------
    2D numpy array
      The convolved texture. The dtype of the output array is the same as the
      input arrays. The value returned is always a newly allocated array, even
      with `iterations=0`, in which case a copy of `texture` will be returned.

    Raises
    ------
    TypeError
      If input arrays' dtypes are mismatched.
    ValueError:
      If non-sensical or unknown values are received.

    Notes
    -----
    All input arrays must have the same dtype, which can be either float32 or
    float64.

    Maximum performance is expected for C order arrays.

    With a kernel.size < 5, uv_mode='polarization' is effectively equivalent to
    uv_mode='velocity'. However, this is still a valid use case, so, no warning
    is emitted.

    It is recommended (but not required) to use odd-sized kernels, so that
    forward and backward passes are balanced.

    Kernels cannot contain non-finite (infinite or NaN) values. Although
    unusual, negative values are allowed.

    No effort is made to avoid propagation of NaNs from the input texture.
    However, streamlines will be terminated whenever a pixel where either u or v
    contains a NaN is encountered.

    Infinite values in any input array are not special cased.

    This function is guaranteed to never mutate any input array, and always
    returns a newly allocated array. Thread-safety is thus trivially guaranteed.
    """
    exceptions: list[Exception] = []
    if iterations < 0:
        exceptions.append(
            ValueError(
                f"Invalid number of iterations: {iterations}\n"
                "Expected a strictly positive integer."
            )
        )

    if uv_mode not in _KNOWN_UV_MODES:
        exceptions.append(
            ValueError(
                f"Invalid uv_mode {uv_mode!r}. Expected one of {_KNOWN_UV_MODES}"
            )
        )

    dtype_error_expectations = (
        f"Expected texture, u, v and kernel with identical dtype, from {_SUPPORTED_DTYPES}. "
        f"Got {texture.dtype=}, {u.dtype=}, {v.dtype=}, {kernel.dtype=}"
    )

    input_dtypes = {arr.dtype for arr in (texture, u, v, kernel)}
    if unsupported_dtypes := input_dtypes.difference(_SUPPORTED_DTYPES):
        exceptions.append(
            TypeError(
                f"Found unsupported data type(s): {list(unsupported_dtypes)}. "
                f"{dtype_error_expectations}"
            )
        )

    if len(input_dtypes) != 1:
        exceptions.append(TypeError(f"Data types mismatch. {dtype_error_expectations}"))

    if texture.ndim != 2:
        exceptions.append(
            ValueError(
                f"Expected a texture with exactly two dimensions. Got {texture.ndim=}"
            )
        )
    if np.any(texture < 0):
        exceptions.append(
            ValueError(
                "Found invalid texture element(s). Expected only positive values."
            )
        )
    if u.shape != texture.shape or v.shape != texture.shape:
        exceptions.append(
            ValueError(
                "Shape mismatch: expected texture, u and v with identical shapes. "
                f"Got {texture.shape=}, {u.shape=}, {v.shape=}"
            )
        )

    if kernel.ndim != 1:
        exceptions.append(
            ValueError(
                f"Expected a kernel with exactly one dimension. Got {kernel.ndim=}"
            )
        )
    if np.any(~np.isfinite(kernel)):
        exceptions.append(ValueError("Found non-finite value(s) in kernel."))

    if len(exceptions) == 1:
        raise exceptions[0]
    elif exceptions:
        raise ExceptionGroup("Invalid inputs were received.", exceptions)

    if iterations == 0:
        return texture.copy()

    input_dtype = texture.dtype
    # mypy ignores can be removed once Python 3.9 is dropped.
    # https://github.com/numpy/numpy/issues/28572
    if input_dtype == np.dtype("float32"):
        return convolve_f32(texture, u, v, kernel, uv_mode, iterations)  # type: ignore[arg-type, return-value, unused-ignore] # pyright: ignore[reportArgumentType, reportReturnType]
    elif input_dtype == np.dtype("float64"):
        return convolve_f64(texture, u, v, kernel, uv_mode, iterations)  # type: ignore[arg-type, return-value, unused-ignore] # pyright: ignore[reportArgumentType, reportReturnType]
    else:
        raise RuntimeError  # pragma: no cover



from typing import Literal

from numpy import dtype, ndarray
from numpy import float32 as f32
from numpy import float64 as f64

def convolve_f32(
    texture: ndarray[tuple[int, int], dtype[f32]],
    u: ndarray[tuple[int, int], dtype[f32]],
    v: ndarray[tuple[int, int], dtype[f32]],
    kernel: ndarray[tuple[int], dtype[f32]],
    uv_mode: Literal["velocity", "polarization"] = "velocity",
    iterations: int = 1,
) -> ndarray[tuple[int, int], dtype[f32]]: ...
def convolve_f64(
    texture: ndarray[tuple[int, int], dtype[f64]],
    u: ndarray[tuple[int, int], dtype[f64]],
    v: ndarray[tuple[int, int], dtype[f64]],
    kernel: ndarray[tuple[int], dtype[f64]],
    uv_mode: Literal["velocity", "polarization"] = "velocity",
    iterations: int = 1,
) -> ndarray[tuple[int, int], dtype[f64]]: ...



from itertools import combinations

import numpy as np
import pytest
from numpy.testing import assert_allclose, assert_array_equal

import rlic

prng = np.random.default_rng(0)

NX = 128


def get_convolve_args(nx=NX, dtype="float64"):
    return (
        prng.random((nx, nx), dtype=dtype),
        prng.random((nx, nx), dtype=dtype),
        prng.random((nx, nx), dtype=dtype),
        np.linspace(0, 1, 11, dtype=dtype),
    )


img, u, v, kernel = get_convolve_args()


def test_no_iterations():
    out = rlic.convolve(img, u, v, kernel=kernel, iterations=0)
    assert_array_equal(out, img)


@pytest.mark.parametrize("dtype", ["float32", "float64"])
def test_single_iteration(dtype):
    img, u, v, kernel = get_convolve_args(dtype=dtype)
    out_impl = rlic.convolve(img, u, v, kernel=kernel)
    out_expl = rlic.convolve(img, u, v, kernel=kernel, iterations=1)
    assert_array_equal(out_impl, out_expl)


def test_multiple_iterations():
    outs = [rlic.convolve(img, u, v, kernel=kernel, iterations=n) for n in range(3)]
    for o1, o2 in combinations(outs, 2):
        assert np.all(o2 != o1)


def test_uv_symmetry():
    out1 = rlic.convolve(img, u, v, kernel=kernel)
    out2 = rlic.convolve(img.T, v.T, u.T, kernel=kernel).T
    assert_array_equal(out2, out1)


def test_uv_mode_default():
    out_vel_impl = rlic.convolve(img, u, v, kernel=kernel)
    out_vel_expl = rlic.convolve(img, u, v, kernel=kernel, uv_mode="velocity")
    assert_array_equal(out_vel_impl, out_vel_expl)


def test_uv_modes_diff():
    kernel = np.ones(5, dtype="float64")
    u0 = np.ones((NX, NX))
    ii = np.broadcast_to(np.arange(NX), (NX, NX))
    u1 = np.where(ii < NX / 2, u0, -u0)
    u2 = np.where(ii < NX / 2, -u0, u0)
    v = np.zeros((NX, NX))

    out_u1_vel = rlic.convolve(img, u1, v, kernel=kernel, uv_mode="velocity")
    out_u2_vel = rlic.convolve(img, u2, v, kernel=kernel, uv_mode="velocity")
    assert_allclose(out_u2_vel, out_u1_vel, atol=1e-14)

    out_u1_pol = rlic.convolve(img, u1, v, kernel=kernel, uv_mode="polarization")
    out_u2_pol = rlic.convolve(img, u2, v, kernel=kernel, uv_mode="polarization")
    assert_allclose(out_u2_pol, out_u1_pol, atol=1e-14)

    diff = out_u2_vel - out_u2_pol
    assert np.ptp(diff) > 1


@pytest.mark.parametrize("kernel_size", [3, 4])
def test_uv_modes_equiv(kernel_size):
    # with a kernel shorter than 5, uv_mode='polarization' doesn't do anything more or
    # different than uv_mode='velocity'
    kernel = np.ones(kernel_size, dtype="float64")
    out_vel = rlic.convolve(img, u, v, kernel=kernel, uv_mode="velocity")
    out_pol = rlic.convolve(img, u, v, kernel=kernel, uv_mode="polarization")
    assert_array_equal(out_pol, out_vel)


def test_uv_mode_polarization_sym():
    NX = 5
    kernel = np.array([1, 1, 1, 1, 1], dtype="float64")
    shape = (NX, NX)
    img = np.eye(NX)
    ZERO = np.zeros(shape, dtype="float64")
    ONE = np.ones(shape, dtype="float64")
    out_u_forward = rlic.convolve(
        img,
        u=ONE,
        v=ZERO,
        kernel=kernel,
        uv_mode="polarization",
    )
    out_u_backward = rlic.convolve(
        img,
        u=-ONE,
        v=ZERO,
        kernel=kernel,
        uv_mode="polarization",
    )
    assert_allclose(out_u_backward, out_u_forward)

    out_v_forward = rlic.convolve(
        img,
        u=ZERO,
        v=ONE,
        kernel=kernel,
        uv_mode="polarization",
    )
    out_v_backward = rlic.convolve(
        img,
        u=ZERO,
        v=-ONE,
        kernel=kernel,
        uv_mode="polarization",
    )
    assert_allclose(out_v_backward, out_v_forward)


@pytest.mark.parametrize("dtype", ["float32", "float64"])
@pytest.mark.parametrize("niterations", [0, 1, 5])
def test_nan_vectors(dtype, niterations):
    img, _, _, kernel = get_convolve_args(dtype=dtype)
    u = v = np.full_like(img, np.nan)

    # streamlines will all be terminated on the first step,
    # but the starting pixel is still to be accumulated, so we expect
    # the output to be identical to the input, to a scaling factor.
    out = rlic.convolve(img, u, v, kernel=kernel, iterations=niterations)
    scaling_factor = out / img
    assert np.ptp(scaling_factor) == 0.0
    assert scaling_factor[0, 0] == kernel[len(kernel) // 2] ** niterations



from itertools import product

import numpy as np
import pytest
from numpy.testing import assert_allclose

import rlic

pytest.importorskip("vectorplot")
from vectorplot.lic_internal import (  # noqa: E402
    line_integral_convolution as reference_impl,
)

NX, NY = SHAPE = (31, 33)
rng = np.random.default_rng(0)
texture = rng.random(SHAPE, dtype="float32")

ONE = np.ones_like(texture)
ZERO = np.zeros_like(texture)

# define velocity components with a sharp divergence
ii = np.broadcast_to(np.arange(NY), SHAPE)
cond = ii < NY / 2
U1 = np.where(cond, -ONE, ONE)

jj = np.broadcast_to(np.atleast_2d(np.arange(NX)).T, SHAPE)
cond = jj < NX / 2
V1 = np.where(cond, -ONE, ONE)

KL = min(NX, NY)
K0 = np.sin(np.arange(KL) * np.pi / KL, dtype="float32")

test_cases = sorted(
    (u, v, k, uv_mode)
    for u, v, k, uv_mode in product(["u0", "u1"], ["v0", "v1"], ["k0"], ["vel", "pol"])
)


@pytest.fixture(params=test_cases)
def known_answer(request):
    field_u, field_v, field_kernel, field_mode = request.param
    if field_u == "u0":
        u = ZERO
    elif field_u == "u1":
        u = U1
    else:
        raise RuntimeError  # pragma: no cover

    if field_v == "v0":
        v = ZERO
    elif field_v == "v1":
        v = V1
    else:
        raise RuntimeError  # pragma: no cover

    if field_kernel == "k0":
        kernel = K0
    else:
        raise RuntimeError  # pragma: no cover

    if field_mode == "vel":
        uv_mode = "velocity"
    elif field_mode == "pol":
        uv_mode = "polarization"
    else:
        raise RuntimeError  # pragma: no cover

    return (
        u,
        v,
        kernel,
        uv_mode,
        reference_impl(u, v, texture, kernel, int(uv_mode == "polarization")),
    )


def test_outputs(known_answer):
    u, v, kernel, uv_mode, expected = known_answer
    out = rlic.convolve(texture, u, v, kernel=kernel, uv_mode=uv_mode)
    assert_allclose(out, expected, rtol=1.5e-7, atol=1e-6)
